__version__ = '0.1.2'
from mrktmix.optimization import optimize_spend
from mrktmix.transformation import apply_apl
from mrktmix.transformation import apply_coef
from mrktmix.transformation import assess_error
from mrktmix.transformation import collapse_date
from mrktmix.transformation import create_base



import mrktmix as mmm


def test_optimize_spend():
    # with out lower bound and upper bound
    optimum_sol = mmm.optimize_spend(10000, [85.0, 58.0, 70.0], [0.6, 0.4, 0.8], constraint='spend', verbosity=0)
    actual = list(*[*optimum_sol["freevariables"].values()])
    expected = [1956.953269488499, 403.28157861028035, 7639.765235633719]
    assert all([a == b for a, b in zip(actual, expected)])

    # with lower bound and upper bound
    optimum_sol = mmm.optimize_spend(
        10000,
        [85.0, 58.0, 70.0],
        [0.6, 0.4, 0.8],
        spend_lower_limit=[10, 500, 1],
        spend_upper_limit=[10000, 500000, 1000],
        revenue_lower_limit=[500, 696.65, 70.00],
        revenue_upper_limit=[21351.03, 11041.19, 17583.21],
        constraint='spend', verbosity=0)
    actual = list(*[*optimum_sol["freevariables"].values()])
    expected = [7839.090441464894, 1160.9084481323118, 999.9999998719591]
    assert all([a == b for a, b in zip(actual, expected)])



import numpy as np
import pandas as pd

from mrktmix.dataprep import mmm_transform as dp


def create_base(variable, date_input, freq, increasing=False, negative=False, periods=1, panel=None):
    """ Create dummy/base variable for modeling

    :param variable: Name of variables
    :type variable: list
    :param date_input: List of string, list and tuble of Dates used in base variable. If input type is str, periods and freq will be used
     to determine length of date. If input type is list, these dates will be used for base. If input type is tuple, the dates in the tuple
     will behave like range for date
    :type date_input: list
    :param freq: list of frequency strings or frequency strings. It can have multiples, e.g. ‘5H’. See here for a list of pandas frequency
     aliases
    :type freq: Union[list, str]
    :param increasing: Base values will in increaseing order if True. If false, it will be constant, defaults to False
    :type increasing: Union[list, bool], optional
    :param negative: Controls the sign for base value, defaults to False
    :type negative: Union[list,bool], optional
    :param periods: Controls number of period dates will be genrated. Will only be applicable when date_input is str, defaults to 1
    :type periods: Union[list, int], optional
    :param panel: Create multiindex with panel on level 0, defaults to None
    :type panel: Union[list,str], optional
    :return: pandas DataFrame with date and panel index
    :rtype: pandas.DataFrame
    """
    if isinstance(variable, (np.ndarray, list)):
        df = pd.DataFrame({"variable": variable, "date_input": date_input, "freq": freq,
                           "increasing": increasing, "negative": negative, "periods": periods, "panel": panel})
        base_df = (df.apply(lambda row: dp.create_base_(row[0],
                                                        row[1],
                                                        row[2],
                                                        increasing=row[3],
                                                        negative=row[4],
                                                        periods=row[5],
                                                        panel=row[6]),
                            axis=1) .set_index(df.iloc[:,
                                                       0].values) .sum(level=0,
                                                                       min_count=1) .transpose())
        base_df.index.freq = None
    else:
        base_df = dp.create_base_(variable, date_input, freq, increasing=increasing, negative=negative, periods=periods, panel=panel)
    return(base_df)


def apply_apl(dframe, dict_apl):
    """ Apply advertisement decay (carry over effect or decay effect), diminishing return and lag on pandas.DataFrame

    :param dframe: DataFrame with marketing or any other activities like spend
    :type dframe: pandas.DataFrame
    :param dict_apl: Dictionary with list of Adstock, carry over effect or decay effect on activity. Key of the dictionary can be name of
        panel or Bool. If key is true, then transformation is applied at panel level. If key is false, then transformation is applied on
        entire dataframe. If trasformation needs to be applied at panel level (name of panel should at level -2 in multiindex row
    :type dict_apl: dictionary with list values. Keys can be string or bool
    :return: Returns input after applying adstock, power and lag transformation
    :rtype: pandas.DataFrame
    """
    if (len(dict_apl) == 1) and (False in dict_apl.keys()):
        all_vars = dict_apl[False]
        df_transformed = pd.concat([dp.apply_apl_series(dframe[var[0]], var[1], var[2], var[3]) for var in all_vars], axis=1)
    elif (len(dict_apl) == 1) and (True in dict_apl.keys()):
        all_vars = dict_apl[True]
        df_grouped = dframe.groupby(dframe.droplevel(-1).index)
        df_transformed = pd.concat([df_grouped[var[0]].transform(
            lambda x: dp.apply_apl_series(x, var[1], var[2], var[3])) for var in all_vars], axis=1)
        df_transformed.columns = pd.MultiIndex.from_tuples(all_vars)
    else:
        df_transformed = pd.DataFrame()
        for panel, all_vars in dict_apl.items():
            df_transformed = pd.concat([df_transformed, pd.concat([dp.apply_apl_series(
                dframe.loc[[panel], var[0]], var[1], var[2], var[3]) for var in all_vars], axis=1)])
    df_transformed.columns.names = ["Variable", "Adstock", "Power", "Lag"]
    return(df_transformed)


def apply_coef(raw_data, coef, dep_series):
    """ Apply coefficient and transformation on raw data

    :param raw_data: modeling dataframe with date index at level -1. If panel is present, it should be at level -2.
    :type raw_data: pandas.DataFrame
    :param coef: Coefficient and parameter to be applied on modeling dataframe. Coefficient should have tuple of variable, adstock,power
        and lag at index level -1. If panel is present in modeling data, then coefficient must have panel information in index at level -2.
    :type coef: pandas.Series
    :param dep_series: Dependent Series will be used to calculate residual. It must be at same level modeling dataframe
    :type dep_series: pandas.Series or None
    :return: Decomposition of dependent series
    :rtype: pandas.DataFrame
    """
    panel_var = (pd.Series(coef.index.get_level_values(-1))
                 .groupby(coef.droplevel(-1).index)
                 .apply(list)
                 .to_dict())
    after_apl = apply_apl(raw_data, panel_var)
    coef2frame = coef.unstack()
    coef2frame.columns = pd.MultiIndex.from_tuples(coef2frame.columns)
    dep_decomposition = after_apl[coef2frame.columns].mul(coef2frame.reindex(after_apl.index, level=0))
    if dep_series is not None:
        dep_decomposition[("Residual", 0, 1, 0)] = dep_series-dep_decomposition.sum(axis=1)
    return(dep_decomposition)


def collapse_date(dep_decompose, date_dict):
    """ Summarise data after collapsing date (index at level -1). Summarization is based on date dictionary given in input.

    :param dep_decompose: data to be summarised. date should be index with level -1.
    :type dep_decompose: pandas.DataFrame
    :param date_dict: dictionary with name and tuples of dates. Dates are inclusive.
    :type date_dict: dictionary
    :return: Summarise after collpasing date
    :rtype: pandas.DataFrame
    """
    all_decomp_smry = []
    for key, val in date_dict.items():
        selected_date = [upp_lim == low_lim for upp_lim, low_lim in zip(
            val[1] >= dep_decompose.index.get_level_values(-1), dep_decompose.index.get_level_values(-1) >= val[0])]
        if dep_decompose.index.nlevels-1:
            decomp_smry = dep_decompose.iloc[selected_date].sum(level=range(0, dep_decompose.index.nlevels-1))
        else:
            decomp_smry = dep_decompose.iloc[selected_date].sum()
        if isinstance(decomp_smry, pd.Series):
            decomp_smry = decomp_smry.rename(key)
        else:
            decomp_smry = decomp_smry.stack(list(range(0, dep_decompose.columns.nlevels))).rename(key)
        if len(all_decomp_smry):
            all_decomp_smry = pd.concat([all_decomp_smry, decomp_smry], axis=1)
        else:
            all_decomp_smry = decomp_smry.to_frame()
    return(all_decomp_smry)


def assess_error(dep_decompose):
    """ Create actual vs predicted with error terms from given resonse decomposition

    :param dep_decompose: data with decomposition of response variable and Residual term
    :type dep_decompose: pandas.DataFrame
    :return: dataframe with actual, predicted, error and error % to measure accury of model
    :rtype: pandas.DataFrame
    """
    dep = dep_decompose.sum(axis=1).rename("Dependent")
    pred = dep_decompose.drop("Residual", axis=1).sum(axis=1).rename("Predicted")
    residual = (dep-pred).rename("Error")
    residual_perc = (residual/dep).rename("Error %")
    return(pd.concat([dep, pred, residual, residual_perc], axis=1))



import numpy as np
from gpkit import Model
from gpkit import SignomialsEnabled
from gpkit import Variable
from gpkit import VectorVariable


def optimize_spend(
        constraint_ceiling,
        coefficient,
        exponent,
        spend_lower_limit=None,
        spend_upper_limit=None,
        revenue_lower_limit=None,
        revenue_upper_limit=None,
        constraint="spend",
        verbosity=0):
    """Optimizes spend based on constraint_ceiling, coefficient and power. constaint_ceiling can be based on spend or revenue which
    can be specified in constraint parameter. If constraint is spend, then optimization will be spend based. It will use total spend
    as budget and tries to maximize total revenue. If constraint is revenue, then optimization will be goal based. It will use total
    revenue and tries to minimize total spend to achieve target revenue. One can also include additional parameters like lower bound
    and upper bound for spend and revenue for optimization.
    Optimization formulation: Total revenue = sum of all {coefficient*spend^(exponent)}. Each coeffient, spend and exponent is assumed
    for every marketing hiearchy.
    Assumption: Coefficient needs to be positive

    :param constraint_ceiling: ceiling or contrains for spend or revenue used for optimization
    :type constraint_ceiling: numeric
    :param coefficient: coeffient from modeling for given marketing hiearchy
    :type coefficient: list of numeric values
    :param exponent: exponent or power for given marketing hiearchy
    :type exponent: list of numeric values
    :param spend_lower_limit: lower bound for spend for optimization
    :type spend_lower_limit: None or list of numeric values
    :param spend_upper_limit: upper bound for spend for optimization
    :type spend_upper_limit: None or list of numeric values
    :param revenue_lower_limit: lower bound for revenue for optimization
    :type revenue_lower_limit: None or list of numeric values
    :param revenue_upper_limit: upper bound for revenue for optimization
    :type revenue_upper_limit: None or list of numeric values
    :param constraint: attribute for ceiling or contrains for optimization. Its value will be spend if objective is spend based
    optimization or value will be revenue if the goal is goal based optimization. By default, its value is spend.
    :type constraint: string
    :param verbosity: Can be used to print details of optimization. By default its value is 0, which will print short description
    of optimization.
    :type verbosity: 0 or 1
    :return: optimized object from gpkit
    :rtype: gpkit.solution_array.SolutionArray
    """

    no_of_optim_params = len(coefficient)
    # convert to coefficent and power to float
    coef = [float(item) for item in coefficient]
    power = [float(item) for item in exponent]
    if spend_lower_limit is None:
        spend_lower_limit = [0] * no_of_optim_params
    if spend_upper_limit is None:
        spend_upper_limit = [float("Inf")] * no_of_optim_params
    if revenue_lower_limit is None:
        revenue_lower_limit = [0] * no_of_optim_params
    if revenue_upper_limit is None:
        revenue_upper_limit = [float("Inf")] * no_of_optim_params

    # Parameters
    if constraint.lower() == 'revenue':
        b = Variable("b", constraint_ceiling, "-", "total revenue")
    else:
        b = Variable("b", constraint_ceiling, "-", "budget")
    s_min = VectorVariable(no_of_optim_params, "s_{min}", spend_lower_limit, "-", "lower bound for spend")
    s_max = VectorVariable(no_of_optim_params, "s_{max}", spend_upper_limit, "-", "upper bound for spend")
    r_min = VectorVariable(no_of_optim_params, "r_{min}", revenue_lower_limit, "-", "lower bound for revenue")
    r_max = VectorVariable(no_of_optim_params, "r_{max}", revenue_upper_limit, "-", "upper bound for revenue")

    # Decision Variable
    s = VectorVariable(no_of_optim_params, "s", "-", "spend")
    r = VectorVariable(no_of_optim_params, "r", "-", "revenue")
    r = coef * np.power(s, power)
    r_inv = VectorVariable(no_of_optim_params, "r_inv", "-", "inverse of individual revenue")
    r_inv = np.power(coef, -1) * np.power(s, np.multiply(-1, np.power(power, -1)))

    # Constraints
    # must enable signomials for subtraction
    if constraint.lower() == 'revenue':
        # goal/revenue based optimization
        with SignomialsEnabled():
            constraints = [s >= s_min, s <= s_max, r >= r_min, r <= r_max, sum(r) >= b]
    else:
        # spend based optimization
        constraints = [s >= s_min, s <= s_max, r >= r_min, r <= r_max, sum(s) <= b]
    # Objective function
    if constraint.lower() == 'revenue':
        # goal/revenue based optimization
        objective = sum(s)
    else:
        # spend based optimization
        objective = sum(r_inv)

    # Formulate the Model
    m = Model(objective, constraints)
    # Solve the Model and print the results table
    if constraint.lower() == 'revenue':
        sol = m.localsolve(verbosity=verbosity)
    else:
        sol = m.solve(verbosity=verbosity)
    return(sol)
